# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyspark.errors import PySparkAttributeError, PySparkNotImplementedError

from congruity.json_conversion import to_json_conversion
from congruity.rdd_adapter import adapt_to_rdd
from congruity.spark_context_adapter import adapt_to_spark_context


def adapt_session_getattr(self, name):
    if name in ["_jsc", "_jconf", "_jvm", "_jsparkSession"]:
        raise PySparkAttributeError(
            error_class="JVM_ATTRIBUTE_NOT_SUPPORTED", message_parameters={"attr_name": name}
        )
    elif name in ["newSession"]:
        raise PySparkNotImplementedError(
            error_class="NOT_IMPLEMENTED", message_parameters={"feature": f"{name}()"}
        )


def monkey_patch_spark():
    import congruity.patch as p
    from pyspark.sql.connect.session import SparkSession

    if p._monkey_patch_complete:
        return
    from pyspark.sql.connect.dataframe import DataFrame

    # Register all the monkey patches here.
    DataFrame.toJSON = to_json_conversion

    # Patch properties.
    setattr(DataFrame, "rdd", property(adapt_to_rdd))
    setattr(DataFrame, "sparkContext", property(adapt_to_spark_context))

    # Spark Session
    setattr(SparkSession, "sparkContext", property(adapt_to_spark_context))
    SparkSession.__getattr__ = adapt_session_getattr

    p._monkey_patch_complete = True
    return



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
_monkey_patch_complete = False
from congruity.helper.patch import monkey_patch_spark

monkey_patch_spark()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

from pyarrow import RecordBatch
from pyspark import Row
from pyspark.cloudpickle import loads


class WrappedIterator(Iterable):
    """This is a helper class that wraps the iterator of RecordBatches as returned by
    mapInArrow and converts it into an iterator of the underlaying values."""

    def __init__(self, i: Iterable[RecordBatch], first_field=False):
        self._first_field = first_field
        self._iter = i
        self._current_batch = None
        self._current_idx = 0
        self._done = False

    def __next__(self):
        if self._current_batch is None or self._current_idx >= len(self._current_batch):
            self._current_idx = 0
            v: RecordBatch = next(self._iter)
            if self._first_field:
                self._current_batch = [loads(x["__bin_field__"]) for x in v.to_pylist()]
            else:
                self._current_batch = [Row(**x) for x in v.to_pylist()]

        result = self._current_batch[self._current_idx]
        self._current_idx += 1
        return result

    def __iter__(self):
        return self



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, TYPE_CHECKING
from pyspark.cloudpickle import dumps

from congruity.rdd_adapter import RDDAdapter

if TYPE_CHECKING:
    from pyspark.sql.connect.dataframe import DataFrame
    from pyspark.sql.connect.session import SparkSession


class SparkContextAdapter:
    _spark: "SparkSession"

    def __init__(self, spark: "SparkSession"):
        self._spark = spark

    def parallelize(self, data: Any, slices: Optional[int] = None) -> "RDDAdapter":
        # TODO - the slices argument is not ideal here.
        # Create the binary DF from the data
        serialized = map(lambda x: dumps(x), data)
        base_df = self._spark.createDataFrame(serialized, RDDAdapter.BIN_SCHEMA)
        if slices is not None:
            base_df = base_df.repartition(slices)
        return RDDAdapter(
            base_df,
            first_field=True,
        )


def adapt_to_spark_context(self: "DataFrame") -> SparkContextAdapter:
    return SparkContextAdapter(self)
